package org.eclipse.jdt.internal.compiler.apt.model;

import org.eclipse.jdt.internal.compiler.lookup.MethodVerifier;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import javax.lang.model.element.TypeElement;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import javax.lang.model.element.TypeParameterElement;
import javax.lang.model.type.TypeMirror;
import org.eclipse.jdt.internal.compiler.ast.Argument;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationHolder;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.lookup.AptBinaryLocalVariableBinding;
import java.util.ArrayList;
import javax.lang.model.element.VariableElement;
import javax.lang.model.element.PackageElement;
import javax.lang.model.element.Modifier;
import java.util.Set;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import javax.lang.model.element.ElementKind;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import java.util.Collections;
import javax.lang.model.element.Element;
import java.util.List;
import javax.lang.model.element.AnnotationValue;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationBinding;
import javax.lang.model.element.ElementVisitor;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.apt.dispatch.BaseProcessingEnvImpl;
import javax.lang.model.element.Name;
import javax.lang.model.element.ExecutableElement;

public class ExecutableElementImpl extends ElementImpl implements ExecutableElement
{
    private Name _name;
    
    ExecutableElementImpl(BaseProcessingEnvImpl env, MethodBinding binding) {
        super(env, binding);
        this._name = null;
    }
    
    @Override
    public <R, P> R accept(ElementVisitor<R, P> v, P p) {
        return v.visitExecutable(this, p);
    }
    
    @Override
    protected AnnotationBinding[] getAnnotationBindings() {
        return ((MethodBinding)this._binding).getAnnotations();
    }
    
    @Override
    public AnnotationValue getDefaultValue() {
        MethodBinding binding = (MethodBinding)this._binding;
        Object defaultValue = binding.getDefaultValue();
        if (defaultValue != null) {
            return new AnnotationMemberValue(this._env, defaultValue, binding);
        }
        return null;
    }
    
    @Override
    public List<? extends Element> getEnclosedElements() {
        return Collections.emptyList();
    }
    
    @Override
    public Element getEnclosingElement() {
        MethodBinding binding = (MethodBinding)this._binding;
        if (binding.declaringClass == null) {
            return null;
        }
        return this._env.getFactory().newElement(binding.declaringClass);
    }
    
    @Override
    public String getFileName() {
        ReferenceBinding dc = ((MethodBinding)this._binding).declaringClass;
        char[] name = dc.getFileName();
        if (name == null) {
            return null;
        }
        return new String(name);
    }
    
    @Override
    public ElementKind getKind() {
        MethodBinding binding = (MethodBinding)this._binding;
        if (binding.isConstructor()) {
            return ElementKind.CONSTRUCTOR;
        }
        if (CharOperation.equals(binding.selector, TypeConstants.CLINIT)) {
            return ElementKind.STATIC_INIT;
        }
        if (CharOperation.equals(binding.selector, TypeConstants.INIT)) {
            return ElementKind.INSTANCE_INIT;
        }
        return ElementKind.METHOD;
    }
    
    @Override
    public Set<Modifier> getModifiers() {
        MethodBinding binding = (MethodBinding)this._binding;
        return Factory.getModifiers(binding.modifiers, this.getKind());
    }
    
    @Override
    PackageElement getPackage() {
        MethodBinding binding = (MethodBinding)this._binding;
        if (binding.declaringClass == null) {
            return null;
        }
        return this._env.getFactory().newPackageElement(binding.declaringClass.fPackage);
    }
    
    @Override
    public List<? extends VariableElement> getParameters() {
        MethodBinding binding = (MethodBinding)this._binding;
        int length = (binding.parameters == null) ? 0 : binding.parameters.length;
        if (length != 0) {
            AbstractMethodDeclaration methodDeclaration = binding.sourceMethod();
            List<VariableElement> params = new ArrayList(length);
            if (methodDeclaration != null) {
                Argument[] arguments;
                for (int length3 = (arguments = methodDeclaration.arguments).length, k = 0; k < length3; ++k) {
                    Argument argument = arguments[k];
                    VariableElement param2 = new VariableElementImpl(this._env, argument.binding);
                    params.add(param2);
                }
            }
            else {
                AnnotationBinding[][] parameterAnnotationBindings = null;
                AnnotationHolder annotationHolder = binding.declaringClass.retrieveAnnotationHolder(binding, false);
                if (annotationHolder != null) {
                    parameterAnnotationBindings = annotationHolder.getParameterAnnotations();
                }
                TypeBinding[] parameters;
                for (int i = 0, length2 = (parameters = binding.parameters).length, j = 0; j < length2; ++i, ++j) {
                    TypeBinding typeBinding = parameters[j];
                    char[] name = (binding.parameterNames.length > i) ? binding.parameterNames[i] : null;
                    if (name == null) {
                        StringBuilder builder = new StringBuilder("arg");
                        builder.append(i);
                        name = String.valueOf(builder).toCharArray();
                    }
                    VariableElement param = new VariableElementImpl(this._env, new AptBinaryLocalVariableBinding(name, typeBinding, 0, (parameterAnnotationBindings != null) ? parameterAnnotationBindings[i] : null, binding));
                    params.add(param);
                }
            }
            return Collections.unmodifiableList(params);
        }
        else {
            return Collections.emptyList();
        }
    }
    
    @Override
    public TypeMirror getReturnType() {
        MethodBinding binding = (MethodBinding)this._binding;
        if (binding.returnType == null) {
            return null;
        }
        return this._env.getFactory().newTypeMirror(binding.returnType);
    }
    
    @Override
    public Name getSimpleName() {
        MethodBinding binding = (MethodBinding)this._binding;
        if (this._name == null) {
            this._name = new NameImpl(binding.selector);
        }
        return this._name;
    }
    
    @Override
    public List<? extends TypeMirror> getThrownTypes() {
        MethodBinding binding = (MethodBinding)this._binding;
        if (binding.thrownExceptions.length == 0) {
            return Collections.emptyList();
        }
        List<TypeMirror> list = new ArrayList(binding.thrownExceptions.length);
        ReferenceBinding[] thrownExceptions;
        for (int length = (thrownExceptions = binding.thrownExceptions).length, i = 0; i < length; ++i) {
            ReferenceBinding exception = thrownExceptions[i];
            list.add(this._env.getFactory().newTypeMirror(exception));
        }
        return list;
    }
    
    @Override
    public List<? extends TypeParameterElement> getTypeParameters() {
        MethodBinding binding = (MethodBinding)this._binding;
        TypeVariableBinding[] variables = binding.typeVariables();
        if (variables.length == 0) {
            return Collections.emptyList();
        }
        List<TypeParameterElement> params = new ArrayList(variables.length);
        for (TypeVariableBinding variable : variables) {
            params.add(this._env.getFactory().newTypeParameterElement(variable, this));
        }
        return Collections.unmodifiableList(params);
    }
    
    @Override
    public boolean hides(Element hidden) {
        if (!(hidden instanceof ExecutableElementImpl)) {
            return false;
        }
        MethodBinding hiderBinding = (MethodBinding)this._binding;
        MethodBinding hiddenBinding = (MethodBinding)((ExecutableElementImpl)hidden)._binding;
        return hiderBinding != hiddenBinding && !hiddenBinding.isPrivate() && hiderBinding.isStatic() && hiddenBinding.isStatic() && CharOperation.equals(hiddenBinding.selector, hiderBinding.selector) && this._env.getLookupEnvironment().methodVerifier().isMethodSubsignature(hiderBinding, hiddenBinding) && hiderBinding.declaringClass.findSuperTypeOriginatingFrom(hiddenBinding.declaringClass) != null;
    }
    
    @Override
    public boolean isVarArgs() {
        return ((MethodBinding)this._binding).isVarargs();
    }
    
    public boolean overrides(ExecutableElement overridden, TypeElement type) {
        MethodBinding overriddenBinding = (MethodBinding)((ExecutableElementImpl)overridden)._binding;
        ReferenceBinding overriderContext = (ReferenceBinding)((TypeElementImpl)type)._binding;
        if ((MethodBinding)this._binding == overriddenBinding || overriddenBinding.isStatic() || overriddenBinding.isPrivate() || ((MethodBinding)this._binding).isStatic()) {
            return false;
        }
        char[] selector = ((MethodBinding)this._binding).selector;
        if (!CharOperation.equals(selector, overriddenBinding.selector)) {
            return false;
        }
        if (overriderContext.findSuperTypeOriginatingFrom(((MethodBinding)this._binding).declaringClass) == null && ((MethodBinding)this._binding).declaringClass.findSuperTypeOriginatingFrom(overriderContext) == null) {
            return false;
        }
        MethodBinding overriderBinding = new MethodBinding((MethodBinding)this._binding, overriderContext);
        if (overriderBinding.isPrivate()) {
            return false;
        }
        TypeBinding match = overriderBinding.declaringClass.findSuperTypeOriginatingFrom(overriddenBinding.declaringClass);
        if (!(match instanceof ReferenceBinding)) {
            return false;
        }
        MethodBinding[] superMethods = ((ReferenceBinding)match).getMethods(selector);
        LookupEnvironment lookupEnvironment = this._env.getLookupEnvironment();
        if (lookupEnvironment == null) {
            return false;
        }
        MethodVerifier methodVerifier = lookupEnvironment.methodVerifier();
        int i = 0;
        int length = superMethods.length;
        while (i < length) {
            if (superMethods[i].original() == overriddenBinding) {
                return methodVerifier.doesMethodOverride(overriderBinding, superMethods[i]);
            }
            ++i;
        }
        return false;
    }
    
    @Override
    public TypeMirror getReceiverType() {
        return this._env.getFactory().getReceiverType((MethodBinding)this._binding);
    }
    
    @Override
    public boolean isDefault() {
        return this._binding != null && ((MethodBinding)this._binding).isDefaultMethod();
    }
}
